.sd-card-img-top {
  width: 33% !important;
  display: block;
  margin-left: auto;
  margin-right: auto;
  margin-top: 10px;
}

.download-div {
  display: flex;
  flex-direction: row;
  justify-content: space-around;
  margin-bottom: 1rem;
}

.download-button {
  display: inline-block;
  padding: 10px 20px;
  background-color: #3498db;
  color: white;
  border: none;
  border-radius: 5px;
  text-align: center;
  text-decoration: none;
  font-size: 16px;
  cursor: pointer;
  transition: background-color 0.3s;
}

.transparent-image > img {
  background-color: unset !important;
}

.download-button:hover {
  background-color: #2980b9;
}

blockquote {
  padding: unset;
  border-left: unset;
}