#if 0
// Copyright (C) 2023-2025 National Center for Atmospheric Research
// SPDX-License-Identifier: Apache-2.0
#endif

#define MUSICA_ERROR_CATEGORY                           "MUSICA Error"
#define MUSICA_ERROR_CODE_SPECIES_NOT_FOUND             1
#define MUSICA_ERROR_CODE_SOLVER_TYPE_NOT_FOUND         2
#define MUSICA_ERROR_CODE_MAPPING_NOT_FOUND             3
#define MUSICA_ERROR_CODE_MAPPING_OPTIONS_UNDEFINED     4
#define MUSICA_ERROR_CODE_UNKNOWN                       5
#define MUSICA_ERROR_CODE_UNSUPPORTED_SOLVER_STATE_PAIR 6

#define MUSICA_ERROR_CATEGORY_PARSING          "MUSICA Parsing"
#define MUSICA_PARSE_PARSING_FAILED            1
#define MUSICA_PARSE_INVALID_CONFIG_FILE       2
#define MUSICA_PARSE_UNSUPPORTED_VERSION       3
#define MUSICA_PARSE_FAILED_TO_CAST_TO_VERSION 4